(*.eval_funs, rulesets, problems and methods concerning polynamials
   authors: Matthias Goldgruber 2003
   (c) due to copyright terms

   use"../IsacKnowledge/Poly.ML";
   use"IsacKnowledge/Poly.ML";
   use"Poly.ML";

   remove_thy"Poly";
   use_thy"IsacKnowledge/Isac";
****************************************************************.*)

(*.****************************************************************
   remark on 'polynomials'
   WN020919
   there are 5 kinds of expanded normalforms:
[1] 'complete polynomial' (Komplettes Polynom), univariate
   a_0 + a_1.x^1 +...+ a_n.x^n   not (a_n = 0)
	        not (a_n = 0), some a_i may be zero (DON'T disappear),
                variables in monomials lexicographically ordered and complete,
                x written as 1*x^1, ...
[2] 'polynomial' (Polynom), univariate and multivariate
   a_0 + a_1.x +...+ a_n.x^n   not (a_n = 0)
   a_0 + a_1.x_1.x_2^n_12...x_m^n_1m +...+  a_n.x_1^n.x_2^n_n2...x_m^n_nm
	        not (a_n = 0), some a_i may be zero (ie. monomials disappear),
                exponents and coefficients equal 1 are not (WN060904.TODO in cancel_p_)shown,
                and variables in monomials are lexicographically ordered  
   examples: [1]: "1 + (-10) * x ^^^ 1 + 25 * x ^^^ 2"
	     [1]: "11 + 0 * x ^^^ 1 + 1 * x ^^^ 2"
	     [2]: "x + (-50) * x ^^^ 3"
	     [2]: "(-1) * x * y ^^^ 2 + 7 * x ^^^ 3"

[3] 'expanded_term' (Ausmultiplizierter Term):
   pull out unary minus to binary minus, 
   as frequently exercised in schools; other conditions for [2] hold however
   examples: "a ^^^ 2 - 2 * a * b + b ^^^ 2"
	     "4 * x ^^^ 2 - 9 * y ^^^ 2"
[4] 'polynomial_in' (Polynom in): 
   polynomial in 1 variable with arbitrary coefficients
   examples: "2 * x + (-50) * x ^^^ 3"                     (poly in x)
	     "(u + v) + (2 * u ^^^ 2) * a + (-u) * a ^^^ 2 (poly in a)
[5] 'expanded_in' (Ausmultiplizierter Termin in): 
   analoguous to [3] with binary minus like [3]
   examples: "2 * x - 50 * x ^^^ 3"                     (expanded in x)
	     "(u + v) + (2 * u ^^^ 2) * a - u * a ^^^ 2 (expanded in a)
*****************************************************************.*)

"******** Poly.ML begin ******************************************";
theory' := overwritel (!theory', [("Poly.thy",Poly.thy)]);


(* is_polyrat_in becomes true, if no bdv is in the denominator of a fraction*)
fun is_polyrat_in t v = 
    let 
	fun coeff_in c v = v mem (vars c);
   	fun finddivide (_ $ _ $ _ $ _) v = raise error("is_polyrat_in:")
	    (* at the moment there is no term like this, but ....*)
	  | finddivide (t as (Const ("HOL.divide",_) $ _ $ b)) v = not(coeff_in b v)
	  | finddivide (_ $ t1 $ t2) v = (finddivide t1 v) orelse (finddivide t2 v)
	  | finddivide (_ $ t1) v = (finddivide t1 v)
	  | finddivide _ _ = false;
     in
	finddivide t v
    end;
    
fun eval_is_polyrat_in _ _ (p as (Const ("Poly.is'_polyrat'_in",_) $ t $ v)) _  =
    if is_polyrat_in t v then 
	Some ((term2str p) ^ " = True",
	      Trueprop $ (mk_equality (p, HOLogic.true_const)))
    else Some ((term2str p) ^ " = True",
	       Trueprop $ (mk_equality (p, HOLogic.false_const)))
  | eval_is_polyrat_in _ _ _ _ = ((*writeln"### nichts matcht";*) None);


local
    (*.a 'c is coefficient of v' if v does NOT occur in c.*)
    fun coeff_in c v = not (v mem (vars c));
    (*
     val v = (term_of o the o (parse thy)) "x";
     val t = (term_of o the o (parse thy)) "1";
     coeff_in t v;
     (*val it = true : bool*)
     val t = (term_of o the o (parse thy)) "a*b+c";
     coeff_in t v;
     (*val it = true : bool*)
     val t = (term_of o the o (parse thy)) "a*x+c";
     coeff_in t v;
     (*val it = false : bool*)
    *)
    (*. a 'monomial t in variable v' is a term t with
      either (1) v NOT existent in t, or (2) v contained in t,
      if (1) then degree 0
      if (2) then v is a factor on the very right, ev. with exponent.*)
    fun factor_right_deg (*case 2*)
    	    (t as Const ("op *",_) $ t1 $ 
    	       (Const ("Atools.pow",_) $ vv $ Free (d,_))) v =
    	if ((vv = v) andalso (coeff_in t1 v)) then Some (int_of_str' d) else None
      | factor_right_deg 
    	    (t as Const ("Atools.pow",_) $ vv $ Free (d,_)) v =
    	if (vv = v) then Some (int_of_str' d) else None
      | factor_right_deg (t as Const ("op *",_) $ t1 $ vv) v = 
    	if ((vv = v) andalso (coeff_in t1 v))then Some 1 else None
      | factor_right_deg vv v =
    	if (vv = v) then Some 1 else None;    
    fun mono_deg_in m v =
    	if coeff_in m v then (*case 1*) Some 0
    	else factor_right_deg m v;
    (*
     val v = (term_of o the o (parse thy)) "x";
     val t = (term_of o the o (parse thy)) "(a*b+c)*x^^^7";
     mono_deg_in t v;
     (*val it = Some 7*)
     val t = (term_of o the o (parse thy)) "x^^^7";
     mono_deg_in t v;
     (*val it = Some 7*)
     val t = (term_of o the o (parse thy)) "(a*b+c)*x";
     mono_deg_in t v;
     (*val it = Some 1*)
     val t = (term_of o the o (parse thy)) "(a*b+x)*x";
     mono_deg_in t v;
     (*val it = None*)
     val t = (term_of o the o (parse thy)) "x";
     mono_deg_in t v;
     (*val it = Some 1*)
     val t = (term_of o the o (parse thy)) "(a*b+c)";
     mono_deg_in t v;
     (*val it = Some 0*)
     val t = (term_of o the o (parse thy)) "ab - (a*b)*x";
     mono_deg_in t v;
     (*val it = None*)
    *)
    fun expand_deg_in t v =
    	let fun edi ~1 ~1 (Const ("op +",_) $ t1 $ t2) =
    		(case mono_deg_in t2 v of (* $ is left associative*)
    		     Some d' => edi d' d' t1
		   | None => None)
    	      | edi ~1 ~1 (Const ("op -",_) $ t1 $ t2) =
    		(case mono_deg_in t2 v of
    		     Some d' => edi d' d' t1
		   | None => None)
    	      | edi d dmax (Const ("op -",_) $ t1 $ t2) =
    		(case mono_deg_in t2 v of
		     (*RL  orelse ((d=0) andalso (d'=0)) need to handle 3+4-...4  +x*)
    		     Some d' => if ((d > d') orelse ((d=0) andalso (d'=0))) then edi d' dmax t1 else None
		   | None => None)
    	      | edi d dmax (Const ("op +",_) $ t1 $ t2) =
    		(case mono_deg_in t2 v of
		     (*RL  orelse ((d=0) andalso (d'=0)) need to handle 3+4-...4  +x*)
    		     Some d' => if ((d > d') orelse ((d=0) andalso (d'=0))) then edi d' dmax t1 else None
		   | None => None)
    	      | edi ~1 ~1 t =
    		(case mono_deg_in t v of
    		     d as Some _ => d
		   | None => None)
    	      | edi d dmax t = (*basecase last*)
    		(case mono_deg_in t v of
    		     Some d' => if ((d > d') orelse ((d=0) andalso (d'=0)))  then Some dmax else None
		   | None => None)
    	in edi ~1 ~1 t end;
    (*
     val v = (term_of o the o (parse thy)) "x";
     val t = (term_of o the o (parse thy)) "a+b";
     expand_deg_in t v;
     (*val it = Some 0*)   
     val t = (term_of o the o (parse thy)) "(a+b)*x";
     expand_deg_in t v;
     (*Some 1*)   
     val t = (term_of o the o (parse thy)) "a*b - (a+b)*x";
     expand_deg_in t v;
     (*Some 1*)   
     val t = (term_of o the o (parse thy)) "a*b + (a-b)*x";
     expand_deg_in t v;
     (*Some 1*)   
     val t = (term_of o the o (parse thy)) "a*b + (a+b)*x + x^^^2";
     expand_deg_in t v;
    *)   
    fun poly_deg_in t v =
    	let fun edi ~1 ~1 (Const ("op +",_) $ t1 $ t2) =
    		(case mono_deg_in t2 v of (* $ is left associative*)
    		     Some d' => edi d' d' t1
		   | None => None)
    	      | edi d dmax (Const ("op +",_) $ t1 $ t2) =
    		(case mono_deg_in t2 v of
 		     (*RL  orelse ((d=0) andalso (d'=0)) need to handle 3+4-...4  +x*)
   		     Some d' => if ((d > d') orelse ((d=0) andalso (d'=0))) then edi d' dmax t1 else None
		   | None => None)
    	      | edi ~1 ~1 t =
    		(case mono_deg_in t v of
    		     d as Some _ => d
		   | None => None)
    	      | edi d dmax t = (*basecase last*)
    		(case mono_deg_in t v of
    		     Some d' => if ((d > d') orelse ((d=0) andalso (d'=0))) then Some dmax else None
		   | None => None)
    	in edi ~1 ~1 t end;
in

fun is_expanded_in t v =
    case expand_deg_in t v of Some _ => true | None => false;
fun is_poly_in t v =
    case poly_deg_in t v of Some _ => true | None => false;
fun has_degree_in t v =
    case expand_deg_in t v of Some d => d | None => ~1;
end;
(*
 val v = (term_of o the o (parse thy)) "x";
 val t = (term_of o the o (parse thy)) "a*b - (a+b)*x + x^^^2";
 has_degree_in t v;
 (*val it = 2*)
 val t = (term_of o the o (parse thy)) "-8 - 2*x + x^^^2";
 has_degree_in t v;
 (*val it = 2*)
 val t = (term_of o the o (parse thy)) "6 + 13*x + 6*x^^^2";
 has_degree_in t v;
 (*val it = 2*)
*)

(*("is_expanded_in", ("Poly.is'_expanded'_in", eval_is_expanded_in ""))*)
fun eval_is_expanded_in _ _ 
	     (p as (Const ("Poly.is'_expanded'_in",_) $ t $ v)) _ =
    if is_expanded_in t v
    then Some ((term2str p) ^ " = True",
	  Trueprop $ (mk_equality (p, HOLogic.true_const)))
    else Some ((term2str p) ^ " = True",
	  Trueprop $ (mk_equality (p, HOLogic.false_const)))
  | eval_is_expanded_in _ _ _ _ = None;
(*
 val t = (term_of o the o (parse thy)) "(-8 - 2*x + x^^^2) is_expanded_in x";
 val Some (id, t') = eval_is_expanded_in 0 0 t 0;
 (*val id = "Poly.is'_expanded'_in (-8 - 2 * x + x ^^^ 2) x = True"*)
 term2str t';
 (*val it = "Poly.is'_expanded'_in (-8 - 2 * x + x ^^^ 2) x = True"*)
*)
(*("is_poly_in", ("Poly.is'_poly'_in", eval_is_poly_in ""))*)
fun eval_is_poly_in _ _ 
	     (p as (Const ("Poly.is'_poly'_in",_) $ t $ v)) _ =
    if is_poly_in t v
    then Some ((term2str p) ^ " = True",
	  Trueprop $ (mk_equality (p, HOLogic.true_const)))
    else Some ((term2str p) ^ " = True",
	  Trueprop $ (mk_equality (p, HOLogic.false_const)))
  | eval_is_poly_in _ _ _ _ = None;
(*
 val t = (term_of o the o (parse thy)) "(8 + 2*x + x^^^2) is_poly_in x";
 val Some (id, t') = eval_is_poly_in 0 0 t 0;
 (*val id = "Poly.is'_poly'_in (8 + 2 * x + x ^^^ 2) x = True"*)
 term2str t';
 (*val it = "Poly.is'_poly'_in (8 + 2 * x + x ^^^ 2) x = True"*)
*)

(*("has_degree_in", ("Poly.has'_degree'_in", eval_has_degree_in ""))*)
fun eval_has_degree_in _ _ 
	     (p as (Const ("Poly.has'_degree'_in",_) $ t $ v)) _ =
    let val d = has_degree_in t v
	val d' = term_of_num HOLogic.realT d
    in Some ((term2str p) ^ " = " ^ (string_of_int d),
	  Trueprop $ (mk_equality (p, d')))
    end
  | eval_has_degree_in _ _ _ _ = None;
(*
> val t = (term_of o the o (parse thy)) "(-8 - 2*x + x^^^2) has_degree_in x";
> val Some (id, t') = eval_has_degree_in 0 0 t 0;
val id = "Poly.has'_degree'_in (-8 - 2 * x + x ^^^ 2) x = 2" : string
> term2str t';
val it = "Poly.has'_degree'_in (-8 - 2 * x + x ^^^ 2) x = 2" : string
*)

(*..*)
val calculate_Poly =
    append_rls "calculate_PolyFIXXXME.not.impl." e_rls
	       [];

(*.for evaluation of conditions in rewrite rules.*)
val Poly_erls = 
    append_rls "Poly_erls" Atools_erls
               [ Calc ("op =",eval_equal "#equal_"),
		 Thm  ("real_unari_minus",num_str real_unari_minus),
                 Calc ("op +",eval_binop "#add_"),
		 Calc ("op -",eval_binop "#sub_"),
		 Calc ("op *",eval_binop "#mult_"),
		 Calc ("Atools.pow" ,eval_binop "#power_")
		 ];

val poly_crls = 
    append_rls "poly_crls" Atools_crls
               [ Calc ("op =",eval_equal "#equal_"),
		 Thm  ("real_unari_minus",num_str real_unari_minus),
                 Calc ("op +",eval_binop "#add_"),
		 Calc ("op -",eval_binop "#sub_"),
		 Calc ("op *",eval_binop "#mult_"),
		 Calc ("Atools.pow" ,eval_binop "#power_")
		 ];


local (*. for make_polynomial .*)

open Term;  (* for type order = EQUAL | LESS | GREATER *)

fun pr_ord EQUAL = "EQUAL"
  | pr_ord LESS  = "LESS"
  | pr_ord GREATER = "GREATER";

fun dest_hd' (Const (a, T)) =                          (* ~ term.ML *)
  (case a of
     "Atools.pow" => ((("|||||||||||||", 0), T), 0)    (*WN greatest string*)
   | _ => (((a, 0), T), 0))
  | dest_hd' (Free (a, T)) = (((a, 0), T), 1)
  | dest_hd' (Var v) = (v, 2)
  | dest_hd' (Bound i) = ((("", i), dummyT), 3)
  | dest_hd' (Abs (_, T, _)) = ((("", 0), T), 4);

fun get_order_pow (t $ (Free(order,_))) = (* RL FIXXXME:geht zufaellig?WN*)
    	(case int_of_str (order) of
	             Some d => d
		   | None   => 0)
  | get_order_pow _ = 0;

fun size_of_term' (Const(str,_) $ t) =
  if "Atools.pow"= str then 1000 + size_of_term' t else 1+size_of_term' t(*WN*)
  | size_of_term' (Abs (_,_,body)) = 1 + size_of_term' body
  | size_of_term' (f$t) = size_of_term' f  +  size_of_term' t
  | size_of_term' _ = 1;

fun term_ord' pr thy (Abs (_, T, t), Abs(_, U, u)) =       (* ~ term.ML *)
      (case term_ord' pr thy (t, u) of EQUAL => typ_ord (T, U) | ord => ord)
  | term_ord' pr thy (t, u) =
      (if pr then 
	 let
	   val (f, ts) = strip_comb t and (g, us) = strip_comb u;
	   val _=writeln("t= f@ts= \""^
	      ((string_of_cterm o cterm_of (sign_of thy)) f)^"\" @ \"["^
	      (commas(map(string_of_cterm o cterm_of (sign_of thy))ts))^"]\"");
	   val _=writeln("u= g@us= \""^
	      ((string_of_cterm o cterm_of (sign_of thy)) g)^"\" @ \"["^
	      (commas(map(string_of_cterm o cterm_of (sign_of thy))us))^"]\"");
	   val _=writeln("size_of_term(t,u)= ("^
	      (string_of_int(size_of_term' t))^", "^
	      (string_of_int(size_of_term' u))^")");
	   val _=writeln("hd_ord(f,g)      = "^((pr_ord o hd_ord)(f,g)));
	   val _=writeln("terms_ord(ts,us) = "^
			   ((pr_ord o terms_ord str false)(ts,us)));
	   val _=writeln("-------");
	 in () end
       else ();
	 case int_ord (size_of_term' t, size_of_term' u) of
	   EQUAL =>
	     let val (f, ts) = strip_comb t and (g, us) = strip_comb u in
	       (case hd_ord (f, g) of EQUAL => (terms_ord str pr) (ts, us) 
	     | ord => ord)
	     end
	 | ord => ord)
and hd_ord (f, g) =                                        (* ~ term.ML *)
  prod_ord (prod_ord indexname_ord typ_ord) int_ord (dest_hd' f, dest_hd' g)
and terms_ord str pr (ts, us) = 
    list_ord (term_ord' pr (assoc_thy "Isac.thy"))(ts, us);
in

fun ord_make_polynomial (pr:bool) thy (_:subst) tu = 
    (term_ord' pr thy(***) tu = LESS );

end;(*local*)


rew_ord' := overwritel (!rew_ord',
[("termlessI", termlessI),
 ("ord_make_polynomial", ord_make_polynomial false thy)
 ]);


val expand =
  Rls{id = "expand", preconds = [], 
      rew_ord = ("dummy_ord", dummy_ord),
      erls = e_rls,srls = Erls,
      calc = [],
      (*asm_thm = [],*)
      rules = [Thm ("real_add_mult_distrib" ,num_str real_add_mult_distrib),
	       (*"(z1.0 + z2.0) * w = z1.0 * w + z2.0 * w"*)
	       Thm ("real_add_mult_distrib2",num_str real_add_mult_distrib2)
	       (*"w * (z1.0 + z2.0) = w * z1.0 + w * z2.0"*)
	       ], scr = EmptyScr}:rls;

(*----------------- Begin: rulesets for make_polynomial_ -----------------
  'rlsIDs' redefined by MG as 'rlsIDs_' 
                                    ^^^*)

val discard_minus_ = 
  Rls{id = "discard_minus_", preconds = [], 
      rew_ord = ("dummy_ord", dummy_ord),
      erls = e_rls,srls = Erls,
      calc = [],
      (*asm_thm = [],*)
      rules = [Thm ("real_diff_minus",num_str real_diff_minus),
	       (*"a - b = a + -1 * b"*)
	       Thm ("sym_real_mult_minus1",num_str (real_mult_minus1 RS sym))
	       (*- ?z = "-1 * ?z"*)
	       ], scr = EmptyScr}:rls;
val expand_poly_ = 
  Rls{id = "expand_poly_", preconds = [], 
      rew_ord = ("dummy_ord", dummy_ord),
      erls = e_rls,srls = Erls,
      calc = [],
      (*asm_thm = [],*)
      rules = [Thm ("real_plus_binom_pow4",num_str real_plus_binom_pow4),
	       (*"(a + b)^^^4 = ... "*)
	       Thm ("real_plus_binom_pow5",num_str real_plus_binom_pow5),
	       (*"(a + b)^^^5 = ... "*)
	       Thm ("real_plus_binom_pow2",num_str real_plus_binom_pow2),
	       (*"(a + b)^^^2 = a^^^2 + 2*a*b + b^^^2"*)
	       Thm ("real_plus_binom_pow3",num_str real_plus_binom_pow3),
	       (*"(a + b)^^^3 = a^^^3 + 3*a^^^2*b + 3*a*b^^^2 + b^^^3" *)
	       Thm ("real_plus_minus_binom1_p_p",num_str real_plus_minus_binom1_p_p),
	       (*"(a + b)*(a + -1 * b) = a^^^2 + -1*b^^^2"*)
	       Thm ("real_plus_minus_binom2_p_p",num_str real_plus_minus_binom2_p_p),
	       (*"(a + -1 * b)*(a + b) = a^^^2 + -1*b^^^2"*)
	      
	       Thm ("real_add_mult_distrib" ,num_str real_add_mult_distrib),
	       (*"(z1.0 + z2.0) * w = z1.0 * w + z2.0 * w"*)
	       Thm ("real_add_mult_distrib2",num_str real_add_mult_distrib2),
	       (*"w * (z1.0 + z2.0) = w * z1.0 + w * z2.0"*)
	       
	       Thm ("realpow_multI", num_str realpow_multI),
	       (*"(r * s) ^^^ n = r ^^^ n * s ^^^ n"*)
	       Thm ("realpow_pow",num_str realpow_pow)
	       (*"(a ^^^ b) ^^^ c = a ^^^ (b * c)"*)
	       ], scr = EmptyScr}:rls;

(*.the expression contains + - * ^ only ?
   this is weaker than 'is_polynomial' !.*)
fun is_polyexp (Free _) = true
  | is_polyexp (Const ("op +",_) $ Free _ $ Free _) = true
  | is_polyexp (Const ("op -",_) $ Free _ $ Free _) = true
  | is_polyexp (Const ("op *",_) $ Free _ $ Free _) = true
  | is_polyexp (Const ("Atools.pow",_) $ Free _ $ Free _) = true
  | is_polyexp (Const ("op +",_) $ t1 $ t2) = 
               ((is_polyexp t1) andalso (is_polyexp t2))
  | is_polyexp (Const ("op -",_) $ t1 $ t2) = 
               ((is_polyexp t1) andalso (is_polyexp t2))
  | is_polyexp (Const ("op *",_) $ t1 $ t2) = 
               ((is_polyexp t1) andalso (is_polyexp t2))
  | is_polyexp (Const ("Atools.pow",_) $ t1 $ t2) = 
               ((is_polyexp t1) andalso (is_polyexp t2))
  | is_polyexp _ = false;

(*("is_polyexp", ("Poly.is'_polyexp", eval_is_polyexp ""))*)
fun eval_is_polyexp (thmid:string) _ 
		       (t as (Const("Poly.is'_polyexp", _) $ arg)) thy = 
    if is_polyexp arg
    then Some (mk_thmid thmid "" 
			((string_of_cterm o cterm_of (sign_of thy)) arg) "", 
	       Trueprop $ (mk_equality (t, HOLogic.true_const)))
    else Some (mk_thmid thmid "" 
			((string_of_cterm o cterm_of (sign_of thy)) arg) "", 
	       Trueprop $ (mk_equality (t, HOLogic.false_const)))
  | eval_is_polyexp _ _ _ _ = None; 

val expand_poly_rat_ = 
  Rls{id = "expand_poly_rat_", preconds = [], 
      rew_ord = ("dummy_ord", dummy_ord),
      erls =  append_rls "e_rls-is_polyexp" e_rls
	        [Calc ("Poly.is'_polyexp", eval_is_polyexp "")
		 ],
      srls = Erls,
      calc = [],
      (*asm_thm = [],*)
      rules = [Thm ("real_plus_binom_pow4_poly",num_str real_plus_binom_pow4_poly),
	       (*"[| a is_polyexp; b is_polyexp |] ==> (a + b)^^^4 = ... "*)
	       Thm ("real_plus_binom_pow5_poly",num_str real_plus_binom_pow5_poly),
	       (*"[| a is_polyexp; b is_polyexp |] ==> (a + b)^^^5 = ... "*)
	       Thm ("real_plus_binom_pow2_poly",num_str real_plus_binom_pow2_poly),
	       (*"[| a is_polyexp; b is_polyexp |] ==>
		            (a + b)^^^2 = a^^^2 + 2*a*b + b^^^2"*)
	       Thm ("real_plus_binom_pow3_poly",num_str real_plus_binom_pow3_poly),
	       (*"[| a is_polyexp; b is_polyexp |] ==> 
			    (a + b)^^^3 = a^^^3 + 3*a^^^2*b + 3*a*b^^^2 + b^^^3" *)
	       Thm ("real_plus_minus_binom1_p_p",num_str real_plus_minus_binom1_p_p),
	       (*"(a + b)*(a + -1 * b) = a^^^2 + -1*b^^^2"*)
	       Thm ("real_plus_minus_binom2_p_p",num_str real_plus_minus_binom2_p_p),
	       (*"(a + -1 * b)*(a + b) = a^^^2 + -1*b^^^2"*)
	      
	       Thm ("real_add_mult_distrib_poly" ,num_str real_add_mult_distrib_poly),
	       (*"w is_polyexp ==> (z1.0 + z2.0) * w = z1.0 * w + z2.0 * w"*)
	       Thm ("real_add_mult_distrib2_poly",num_str real_add_mult_distrib2_poly),
	       (*"w is_polyexp ==> w * (z1.0 + z2.0) = w * z1.0 + w * z2.0"*)
	       
	       Thm ("realpow_multI_poly", num_str realpow_multI_poly),
	       (*"[| r is_polyexp; s is_polyexp |] ==> 
		            (r * s) ^^^ n = r ^^^ n * s ^^^ n"*)
	       Thm ("realpow_pow",num_str realpow_pow)
	       (*"(a ^^^ b) ^^^ c = a ^^^ (b * c)"*)
	       ], scr = EmptyScr}:rls;

val simplify_power_ = 
  Rls{id = "simplify_power_", preconds = [], 
      rew_ord = ("dummy_ord", dummy_ord),
      erls = e_rls, srls = Erls,
      calc = [],
      (*asm_thm = [],*)
      rules = [(*MG: Reihenfolge der folgenden 2 Thm muss so bleiben, wegen
		a*(a*a) --> a*a^^^2 und nicht a*(a*a) --> a^^^2*a *)
	       Thm ("sym_realpow_twoI",num_str (realpow_twoI RS sym)),	
	       (*"r * r = r ^^^ 2"*)
	       Thm ("realpow_twoI_assoc_l",num_str realpow_twoI_assoc_l),
	       (*"r * (r * s) = r ^^^ 2 * s"*)

	       Thm ("realpow_plus_1",num_str realpow_plus_1),		
	       (*"r * r ^^^ n = r ^^^ (n + 1)"*)
	       Thm ("realpow_plus_1_assoc_l", num_str realpow_plus_1_assoc_l),
	       (*"r * (r ^^^ m * s) = r ^^^ (1 + m) * s"*)
	       (*MG 9.7.03: neues Thm wegen a*(a*(a*b)) --> a^^^2*(a*b) *)
	       Thm ("realpow_plus_1_assoc_l2", num_str realpow_plus_1_assoc_l2),
	       (*"r ^^^ m * (r * s) = r ^^^ (1 + m) * s"*)

	       Thm ("sym_realpow_addI",num_str (realpow_addI RS sym)),
	       (*"r ^^^ n * r ^^^ m = r ^^^ (n + m)"*)
	       Thm ("realpow_addI_assoc_l", num_str realpow_addI_assoc_l),
	       (*"r ^^^ n * (r ^^^ m * s) = r ^^^ (n + m) * s"*)
	       
	       (* ist in expand_poly - wird hier aber auch gebraucht, wegen: 
		  "r * r = r ^^^ 2" wenn r=a^^^b*)
	       Thm ("realpow_pow",num_str realpow_pow)
	       (*"(a ^^^ b) ^^^ c = a ^^^ (b * c)"*)
	       ], scr = EmptyScr}:rls;

val calc_add_mult_pow_ = 
  Rls{id = "calc_add_mult_pow_", preconds = [], 
      rew_ord = ("dummy_ord", dummy_ord),
      erls = Atools_erls(*erls3.4.03*),srls = Erls,
      calc = [("plus"  , ("op +", eval_binop "#add_")), 
	      ("times" , ("op *", eval_binop "#mult_")),
	      ("power_", ("Atools.pow", eval_binop "#power_"))
	      ],
      (*asm_thm = [],*)
      rules = [Calc ("op +", eval_binop "#add_"),
	       Calc ("op *", eval_binop "#mult_"),
	       Calc ("Atools.pow", eval_binop "#power_")
	       ], scr = EmptyScr}:rls;

val reduce_012_mult_ = 
  Rls{id = "reduce_012_mult_", preconds = [], 
      rew_ord = ("dummy_ord", dummy_ord),
      erls = e_rls,srls = Erls,
      calc = [],
      (*asm_thm = [],*)
      rules = [(* MG: folgende Thm mssen hier stehen bleiben: *)
               Thm ("real_mult_1_right",num_str real_mult_1_right),
	       (*"z * 1 = z"*) (*wegen "a * b * b^^^(-1) + a"*) 
	       Thm ("realpow_zeroI",num_str realpow_zeroI),
	       (*"r ^^^ 0 = 1"*) (*wegen "a*a^^^(-1)*c + b + c"*)
	       Thm ("realpow_oneI",num_str realpow_oneI),
	       (*"r ^^^ 1 = r"*)
	       Thm ("realpow_eq_oneI",num_str realpow_eq_oneI)
	       (*"1 ^^^ n = 1"*)
	       ], scr = EmptyScr}:rls;

val collect_numerals_ = 
  Rls{id = "collect_numerals_", preconds = [], 
      rew_ord = ("dummy_ord", dummy_ord),
      erls = Atools_erls(*erls3.4.03*),srls = Erls,
      calc = [("plus"  , ("op +", eval_binop "#add_"))
	      ],
      (*asm_thm = [],*)
      rules = [Thm ("real_num_collect",num_str real_num_collect), 
	       (*"[| l is_const; m is_const |]==>l * n + m * n = (l + m) * n"*)
	       Thm ("real_num_collect_assoc_r",num_str real_num_collect_assoc_r),
	       (*"[| l is_const; m is_const |] ==>  \
					\(k + m * n) + l * n = k + (l + m) * n"*)
	       Thm ("real_one_collect",num_str real_one_collect),	
	       (*"m is_const ==> n + m * n = (1 + m) * n"*)
	       Thm ("real_one_collect_assoc_r",num_str real_one_collect_assoc_r), 
	       (*"m is_const ==> (k + n) + m * n = k + (m + 1) * n"*)

	 	Calc ("op +", eval_binop "#add_"),

	       (*MG: Reihenfolge der folgenden 2 Thm muss so bleiben, wegen
		     (a+a)+a --> a + 2*a --> 3*a und nicht (a+a)+a --> 2*a + a *)
		Thm ("real_mult_2_assoc_r",num_str real_mult_2_assoc_r),
	       (*"(k + z1) + z1 = k + 2 * z1"*)
	       Thm ("sym_real_mult_2",num_str (real_mult_2 RS sym))
	       (*"z1 + z1 = 2 * z1"*)
	       
	       ], scr = EmptyScr}:rls;

val reduce_012_ = 
  Rls{id = "reduce_012_", preconds = [], 
      rew_ord = ("dummy_ord", dummy_ord),
      erls = e_rls,srls = Erls,
      calc = [],
      (*asm_thm = [],*)
      rules = [Thm ("real_mult_1",num_str real_mult_1),                 
	       (*"1 * z = z"*)
	       Thm ("real_mult_0",num_str real_mult_0),        
	       (*"0 * z = 0"*)
	       Thm ("real_add_zero_left",num_str real_add_zero_left),
	       (*"0 + z = z"*)
	       Thm ("real_add_zero_right",num_str real_add_zero_right),
	       (*"z + 0 = z"*) (*wegen a+b-b --> a+(1-1)*b --> a+0 --> a*)

	       (*Thm ("realpow_oneI",num_str realpow_oneI)*)
	       (*"?r ^^^ 1 = ?r"*)
	       Thm ("real_0_divide",num_str real_0_divide)(*WN060914*)
	       (*"0 / ?x = 0"*)
	       ], scr = EmptyScr}:rls;

(*ein Hilfs-'ruleset' (benutzt das leere 'ruleset')*)
val discard_parentheses_ = 
    append_rls "discard_parentheses_" e_rls 
	       [Thm ("sym_real_mult_assoc", num_str (real_mult_assoc RS sym))
		(*"?z1.1 * (?z2.1 * ?z3.1) = ?z1.1 * ?z2.1 * ?z3.1"*)
		(*Thm ("sym_real_add_assoc",num_str (real_add_assoc RS sym))*)
		(*"?z1.1 + (?z2.1 + ?z3.1) = ?z1.1 + ?z2.1 + ?z3.1"*)
		 ];

(*----------------- End: rulesets for make_polynomial_ -----------------*)

(*MG.0401 ev. for use in rls with ordered rewriting ?
val collect_numerals_left = 
  Rls{id = "collect_numerals", preconds = [], 
      rew_ord = ("dummy_ord", dummy_ord),
      erls = Atools_erls(*erls3.4.03*),srls = Erls,
      calc = [("plus"  , ("op +", eval_binop "#add_")), 
	      ("times" , ("op *", eval_binop "#mult_")),
	      ("power_", ("Atools.pow", eval_binop "#power_"))
	      ],
      (*asm_thm = [],*)
      rules = [Thm ("real_num_collect",num_str real_num_collect), 
	       (*"[| l is_const; m is_const |]==>l * n + m * n = (l + m) * n"*)
	       Thm ("real_num_collect_assoc",num_str real_num_collect_assoc),
	       (*"[| l is_const; m is_const |] ==>  
				l * n + (m * n + k) =  (l + m) * n + k"*)
	       Thm ("real_one_collect",num_str real_one_collect),	
	       (*"m is_const ==> n + m * n = (1 + m) * n"*)
	       Thm ("real_one_collect_assoc",num_str real_one_collect_assoc), 
	       (*"m is_const ==> n + (m * n + k) = (1 + m) * n + k"*)
	       
	       Calc ("op +", eval_binop "#add_"),

	       (*MG am 2.5.03: 2 Theoreme aus reduce_012 hierher verschoben*)
	       Thm ("sym_real_mult_2",num_str (real_mult_2 RS sym)),	
	       (*"z1 + z1 = 2 * z1"*)
	       Thm ("real_mult_2_assoc",num_str real_mult_2_assoc)
	       (*"z1 + (z1 + k) = 2 * z1 + k"*)
	       ], scr = EmptyScr}:rls;*)

val expand_poly = 
  Rls{id = "expand_poly", preconds = [], 
      rew_ord = ("dummy_ord", dummy_ord),
      erls = e_rls,srls = Erls,
      calc = [],
      (*asm_thm = [],*)
      rules = [Thm ("real_add_mult_distrib" ,num_str real_add_mult_distrib),
	       (*"(z1.0 + z2.0) * w = z1.0 * w + z2.0 * w"*)
	       Thm ("real_add_mult_distrib2",num_str real_add_mult_distrib2),
	       (*"w * (z1.0 + z2.0) = w * z1.0 + w * z2.0"*)
	       (*Thm ("real_add_mult_distrib1",num_str real_add_mult_distrib1),
		....... 18.3.03 undefined???*)

	       Thm ("real_plus_binom_pow2",num_str real_plus_binom_pow2),
	       (*"(a + b)^^^2 = a^^^2 + 2*a*b + b^^^2"*)
	       Thm ("real_minus_binom_pow2_p",num_str real_minus_binom_pow2_p),
	       (*"(a - b)^^^2 = a^^^2 + -2*a*b + b^^^2"*)
	       Thm ("real_plus_minus_binom1_p",
		    num_str real_plus_minus_binom1_p),
	       (*"(a + b)*(a - b) = a^^^2 + -1*b^^^2"*)
	       Thm ("real_plus_minus_binom2_p",
		    num_str real_plus_minus_binom2_p),
	       (*"(a - b)*(a + b) = a^^^2 + -1*b^^^2"*)

	       Thm ("real_minus_minus",num_str real_minus_minus),
	       (*"- (- ?z) = ?z"*)
	       Thm ("real_diff_minus",num_str real_diff_minus),
	       (*"a - b = a + -1 * b"*)
	       Thm ("sym_real_mult_minus1",num_str (real_mult_minus1 RS sym))
	       (*- ?z = "-1 * ?z"*)

	       (*Thm ("",num_str ),
	       Thm ("",num_str ),
	       Thm ("",num_str ),*)
	       (*Thm ("real_minus_add_distrib",
		      num_str real_minus_add_distrib),*)
	       (*"- (?x + ?y) = - ?x + - ?y"*)
	       (*Thm ("real_diff_plus",num_str real_diff_plus)*)
	       (*"a - b = a + -b"*)
	       ], scr = EmptyScr}:rls;
val simplify_power = 
  Rls{id = "simplify_power", preconds = [], 
      rew_ord = ("dummy_ord", dummy_ord),
      erls = e_rls, srls = Erls,
      calc = [],
      (*asm_thm = [],*)
      rules = [Thm ("realpow_multI", num_str realpow_multI),
	       (*"(r * s) ^^^ n = r ^^^ n * s ^^^ n"*)
	       
	       Thm ("sym_realpow_twoI",num_str (realpow_twoI RS sym)),	
	       (*"r1 * r1 = r1 ^^^ 2"*)
	       Thm ("realpow_plus_1",num_str realpow_plus_1),		
	       (*"r * r ^^^ n = r ^^^ (n + 1)"*)
	       Thm ("realpow_pow",num_str realpow_pow),
	       (*"(a ^^^ b) ^^^ c = a ^^^ (b * c)"*)
	       Thm ("sym_realpow_addI",num_str (realpow_addI RS sym)),
	       (*"r ^^^ n * r ^^^ m = r ^^^ (n + m)"*)
	       Thm ("realpow_oneI",num_str realpow_oneI),
	       (*"r ^^^ 1 = r"*)
	       Thm ("realpow_eq_oneI",num_str realpow_eq_oneI)
	       (*"1 ^^^ n = 1"*)
	       ], scr = EmptyScr}:rls;
(*MG.0401: termorders for multivariate polys dropped due to principal problems:
  (total-degree-)ordering of monoms NOT possible with size_of_term GIVEN*)
val order_add_mult = 
  Rls{id = "order_add_mult", preconds = [], 
      rew_ord = ("ord_make_polynomial",ord_make_polynomial false Poly.thy),
      erls = e_rls,srls = Erls,
      calc = [],
      (*asm_thm = [],*)
      rules = [Thm ("real_mult_commute",num_str real_mult_commute),
	       (* z * w = w * z *)
	       Thm ("real_mult_left_commute",num_str real_mult_left_commute),
	       (*z1.0 * (z2.0 * z3.0) = z2.0 * (z1.0 * z3.0)*)
	       Thm ("real_mult_assoc",num_str real_mult_assoc),		
	       (*z1.0 * z2.0 * z3.0 = z1.0 * (z2.0 * z3.0)*)
	       Thm ("real_add_commute",num_str real_add_commute),	
	       (*z + w = w + z*)
	       Thm ("real_add_left_commute",num_str real_add_left_commute),
	       (*x + (y + z) = y + (x + z)*)
	       Thm ("real_add_assoc",num_str real_add_assoc)	               
	       (*z1.0 + z2.0 + z3.0 = z1.0 + (z2.0 + z3.0)*)
	       ], scr = EmptyScr}:rls;
(*MG.0401: termorders for multivariate polys dropped due to principal problems:
  (total-degree-)ordering of monoms NOT possible with size_of_term GIVEN*)
val order_mult = 
  Rls{id = "order_mult", preconds = [], 
      rew_ord = ("ord_make_polynomial",ord_make_polynomial false Poly.thy),
      erls = e_rls,srls = Erls,
      calc = [],
      (*asm_thm = [],*)
      rules = [Thm ("real_mult_commute",num_str real_mult_commute),
	       (* z * w = w * z *)
	       Thm ("real_mult_left_commute",num_str real_mult_left_commute),
	       (*z1.0 * (z2.0 * z3.0) = z2.0 * (z1.0 * z3.0)*)
	       Thm ("real_mult_assoc",num_str real_mult_assoc)	
	       (*z1.0 * z2.0 * z3.0 = z1.0 * (z2.0 * z3.0)*)
	       ], scr = EmptyScr}:rls;
val collect_numerals = 
  Rls{id = "collect_numerals", preconds = [], 
      rew_ord = ("dummy_ord", dummy_ord),
      erls = Atools_erls(*erls3.4.03*),srls = Erls,
      calc = [("plus"  , ("op +", eval_binop "#add_")), 
	      ("times" , ("op *", eval_binop "#mult_")),
	      ("power_", ("Atools.pow", eval_binop "#power_"))
	      ],
      (*asm_thm = [],*)
      rules = [Thm ("real_num_collect",num_str real_num_collect), 
	       (*"[| l is_const; m is_const |]==>l * n + m * n = (l + m) * n"*)
	       Thm ("real_num_collect_assoc",num_str real_num_collect_assoc),
	       (*"[| l is_const; m is_const |] ==>  
				l * n + (m * n + k) =  (l + m) * n + k"*)
	       Thm ("real_one_collect",num_str real_one_collect),	
	       (*"m is_const ==> n + m * n = (1 + m) * n"*)
	       Thm ("real_one_collect_assoc",num_str real_one_collect_assoc), 
	       (*"m is_const ==> k + (n + m * n) = k + (1 + m) * n"*)
	       Calc ("op +", eval_binop "#add_"), 
	       Calc ("op *", eval_binop "#mult_"),
	       Calc ("Atools.pow", eval_binop "#power_")
	       ], scr = EmptyScr}:rls;
val reduce_012 = 
  Rls{id = "reduce_012", preconds = [], 
      rew_ord = ("dummy_ord", dummy_ord),
      erls = e_rls,srls = Erls,
      calc = [],
      (*asm_thm = [],*)
      rules = [Thm ("real_mult_1",num_str real_mult_1),                 
	       (*"1 * z = z"*)
	       (*Thm ("real_mult_minus1",num_str real_mult_minus1),14.3.03*)
	       (*"-1 * z = - z"*)
	       Thm ("sym_real_mult_minus_eq1", 
		    num_str (real_mult_minus_eq1 RS sym)),
	       (*- (?x * ?y) = "- ?x * ?y"*)
	       (*Thm ("real_minus_mult_cancel",num_str real_minus_mult_cancel),
	       (*"- ?x * - ?y = ?x * ?y"*)---*)
	       Thm ("real_mult_0",num_str real_mult_0),        
	       (*"0 * z = 0"*)
	       Thm ("real_add_zero_left",num_str real_add_zero_left),
	       (*"0 + z = z"*)
	       Thm ("real_add_minus",num_str real_add_minus),
	       (*"?z + - ?z = 0"*)
	       Thm ("sym_real_mult_2",num_str (real_mult_2 RS sym)),	
	       (*"z1 + z1 = 2 * z1"*)
	       Thm ("real_mult_2_assoc",num_str real_mult_2_assoc)
	       (*"z1 + (z1 + k) = 2 * z1 + k"*)
	       ], scr = EmptyScr}:rls;
(*ein Hilfs-'ruleset' (benutzt das leere 'ruleset')*)
val discard_parentheses = 
    append_rls "discard_parentheses" e_rls 
	       [Thm ("sym_real_mult_assoc", num_str (real_mult_assoc RS sym)),
		Thm ("sym_real_add_assoc",num_str (real_add_assoc RS sym))];

val scr_make_polynomial = 
"Script Expand_binoms t_ =\
\(Repeat                       \
\((Try (Repeat (Rewrite real_diff_minus         False))) @@ \ 

\ (Try (Repeat (Rewrite real_add_mult_distrib   False))) @@ \	 
\ (Try (Repeat (Rewrite real_add_mult_distrib2  False))) @@ \	
\ (Try (Repeat (Rewrite real_diff_mult_distrib  False))) @@ \	
\ (Try (Repeat (Rewrite real_diff_mult_distrib2 False))) @@ \	

\ (Try (Repeat (Rewrite real_mult_1             False))) @@ \		   
\ (Try (Repeat (Rewrite real_mult_0             False))) @@ \		   
\ (Try (Repeat (Rewrite real_add_zero_left      False))) @@ \	 

\ (Try (Repeat (Rewrite real_mult_commute       False))) @@ \		
\ (Try (Repeat (Rewrite real_mult_left_commute  False))) @@ \	
\ (Try (Repeat (Rewrite real_mult_assoc         False))) @@ \		
\ (Try (Repeat (Rewrite real_add_commute        False))) @@ \		
\ (Try (Repeat (Rewrite real_add_left_commute   False))) @@ \	 
\ (Try (Repeat (Rewrite real_add_assoc          False))) @@ \	 

\ (Try (Repeat (Rewrite sym_realpow_twoI        False))) @@ \	 
\ (Try (Repeat (Rewrite realpow_plus_1          False))) @@ \	 
\ (Try (Repeat (Rewrite sym_real_mult_2         False))) @@ \		
\ (Try (Repeat (Rewrite real_mult_2_assoc       False))) @@ \		

\ (Try (Repeat (Rewrite real_num_collect        False))) @@ \		
\ (Try (Repeat (Rewrite real_num_collect_assoc  False))) @@ \	

\ (Try (Repeat (Rewrite real_one_collect        False))) @@ \		
\ (Try (Repeat (Rewrite real_one_collect_assoc  False))) @@ \   

\ (Try (Repeat (Calculate plus  ))) @@ \
\ (Try (Repeat (Calculate times ))) @@ \
\ (Try (Repeat (Calculate power_)))) \  
\ t_)";

(*version used by MG.02/03, overwritten by version AG in 04 below
val make_polynomial = prep_rls(
  Seq{id = "make_polynomial", preconds = []:term list, 
      rew_ord = ("dummy_ord", dummy_ord),
      erls = Atools_erls, srls = Erls,
      calc = [],(*asm_thm = [],*)
      rules = [Rls_ expand_poly,
	       Rls_ order_add_mult,
	       Rls_ simplify_power,   (*realpow_eq_oneI, eg. x^1 --> x *)
	       Rls_ collect_numerals, (*eg. x^(2+ -1) --> x^1          *)
	       Rls_ reduce_012,
	       Thm ("realpow_oneI",num_str realpow_oneI),(*in --^*) 
	       Rls_ discard_parentheses
	       ],
      scr = EmptyScr
      }:rls);   *)

val scr_expand_binoms =
"Script Expand_binoms t_ =\
\(Repeat                       \
\((Try (Repeat (Rewrite real_plus_binom_pow2    False))) @@ \
\ (Try (Repeat (Rewrite real_plus_binom_times   False))) @@ \
\ (Try (Repeat (Rewrite real_minus_binom_pow2   False))) @@ \
\ (Try (Repeat (Rewrite real_minus_binom_times  False))) @@ \
\ (Try (Repeat (Rewrite real_plus_minus_binom1  False))) @@ \
\ (Try (Repeat (Rewrite real_plus_minus_binom2  False))) @@ \

\ (Try (Repeat (Rewrite real_mult_1             False))) @@ \
\ (Try (Repeat (Rewrite real_mult_0             False))) @@ \
\ (Try (Repeat (Rewrite real_add_zero_left      False))) @@ \

\ (Try (Repeat (Calculate plus  ))) @@ \
\ (Try (Repeat (Calculate times ))) @@ \
\ (Try (Repeat (Calculate power_))) @@ \

\ (Try (Repeat (Rewrite sym_realpow_twoI        False))) @@ \
\ (Try (Repeat (Rewrite realpow_plus_1          False))) @@ \
\ (Try (Repeat (Rewrite sym_real_mult_2         False))) @@ \
\ (Try (Repeat (Rewrite real_mult_2_assoc       False))) @@ \

\ (Try (Repeat (Rewrite real_num_collect        False))) @@ \
\ (Try (Repeat (Rewrite real_num_collect_assoc  False))) @@ \

\ (Try (Repeat (Rewrite real_one_collect        False))) @@ \
\ (Try (Repeat (Rewrite real_one_collect_assoc  False))) @@ \ 

\ (Try (Repeat (Calculate plus  ))) @@ \
\ (Try (Repeat (Calculate times ))) @@ \
\ (Try (Repeat (Calculate power_)))) \  
\ t_)";

val expand_binoms = 
  Rls{id = "expand_binoms", preconds = [], rew_ord = ("termlessI",termlessI),
      erls = Atools_erls, srls = Erls,
      calc = [("plus"  , ("op +", eval_binop "#add_")), 
	      ("times" , ("op *", eval_binop "#mult_")),
	      ("power_", ("Atools.pow", eval_binop "#power_"))
	      ],
      (*asm_thm = [],*)
      rules = [Thm ("real_plus_binom_pow2"  ,num_str real_plus_binom_pow2),     
	       (*"(a + b) ^^^ 2 = a ^^^ 2 + 2 * a * b + b ^^^ 2"*)
	       Thm ("real_plus_binom_times" ,num_str real_plus_binom_times),    
	      (*"(a + b)*(a + b) = ...*)
	       Thm ("real_minus_binom_pow2" ,num_str real_minus_binom_pow2),   
	       (*"(a - b) ^^^ 2 = a ^^^ 2 - 2 * a * b + b ^^^ 2"*)
	       Thm ("real_minus_binom_times",num_str real_minus_binom_times),   
	       (*"(a - b)*(a - b) = ...*)
	       Thm ("real_plus_minus_binom1",num_str real_plus_minus_binom1),   
		(*"(a + b) * (a - b) = a ^^^ 2 - b ^^^ 2"*)
	       Thm ("real_plus_minus_binom2",num_str real_plus_minus_binom2),   
		(*"(a - b) * (a + b) = a ^^^ 2 - b ^^^ 2"*)
	       (*RL 020915*)
	       Thm ("real_pp_binom_times",num_str real_pp_binom_times), 
		(*(a + b)*(c + d) = a*c + a*d + b*c + b*d*)
               Thm ("real_pm_binom_times",num_str real_pm_binom_times), 
		(*(a + b)*(c - d) = a*c - a*d + b*c - b*d*)
               Thm ("real_mp_binom_times",num_str real_mp_binom_times), 
		(*(a - b)*(c + d) = a*c + a*d - b*c - b*d*)
               Thm ("real_mm_binom_times",num_str real_mm_binom_times), 
		(*(a - b)*(c - d) = a*c - a*d - b*c + b*d*)
	       Thm ("realpow_multI",num_str realpow_multI),                
		(*(a*b)^^^n = a^^^n * b^^^n*)
	       Thm ("real_plus_binom_pow3",num_str real_plus_binom_pow3),
	        (* (a + b)^^^3 = a^^^3 + 3*a^^^2*b + 3*a*b^^^2 + b^^^3 *)
	       Thm ("real_minus_binom_pow3",num_str real_minus_binom_pow3),
	        (* (a - b)^^^3 = a^^^3 - 3*a^^^2*b + 3*a*b^^^2 - b^^^3 *)


             (*  Thm ("real_add_mult_distrib" ,num_str real_add_mult_distrib),	
		(*"(z1.0 + z2.0) * w = z1.0 * w + z2.0 * w"*)
	       Thm ("real_add_mult_distrib2",num_str real_add_mult_distrib2),	
	       (*"w * (z1.0 + z2.0) = w * z1.0 + w * z2.0"*)
	       Thm ("real_diff_mult_distrib" ,num_str real_diff_mult_distrib),	
	       (*"(z1.0 - z2.0) * w = z1.0 * w - z2.0 * w"*)
	       Thm ("real_diff_mult_distrib2",num_str real_diff_mult_distrib2),	
	       (*"w * (z1.0 - z2.0) = w * z1.0 - w * z2.0"*)
	       *)
	       
	       Thm ("real_mult_1",num_str real_mult_1),              (*"1 * z = z"*)
	       Thm ("real_mult_0",num_str real_mult_0),              (*"0 * z = 0"*)
	       Thm ("real_add_zero_left",num_str real_add_zero_left),(*"0 + z = z"*)

	       Calc ("op +", eval_binop "#add_"), 
	       Calc ("op *", eval_binop "#mult_"),
	       Calc ("Atools.pow", eval_binop "#power_"),
               (*	       
	        Thm ("real_mult_commute",num_str real_mult_commute),		(*AC-rewriting*)
	       Thm ("real_mult_left_commute",num_str real_mult_left_commute),	(**)
	       Thm ("real_mult_assoc",num_str real_mult_assoc),			(**)
	       Thm ("real_add_commute",num_str real_add_commute),		(**)
	       Thm ("real_add_left_commute",num_str real_add_left_commute),	(**)
	       Thm ("real_add_assoc",num_str real_add_assoc),	                (**)
	       *)
	       
	       Thm ("sym_realpow_twoI",num_str (realpow_twoI RS sym)),		
	       (*"r1 * r1 = r1 ^^^ 2"*)
	       Thm ("realpow_plus_1",num_str realpow_plus_1),			
	       (*"r * r ^^^ n = r ^^^ (n + 1)"*)
	       (*Thm ("sym_real_mult_2",num_str (real_mult_2 RS sym)),		
	       (*"z1 + z1 = 2 * z1"*)*)
	       Thm ("real_mult_2_assoc",num_str real_mult_2_assoc),		
	       (*"z1 + (z1 + k) = 2 * z1 + k"*)

	       Thm ("real_num_collect",num_str real_num_collect), 
	       (*"[| l is_const; m is_const |] ==> l * n + m * n = (l + m) * n"*)
	       Thm ("real_num_collect_assoc",num_str real_num_collect_assoc),	
	       (*"[| l is_const; m is_const |] ==>  l * n + (m * n + k) =  (l + m) * n + k"*)
	       Thm ("real_one_collect",num_str real_one_collect),		
	       (*"m is_const ==> n + m * n = (1 + m) * n"*)
	       Thm ("real_one_collect_assoc",num_str real_one_collect_assoc), 
	       (*"m is_const ==> k + (n + m * n) = k + (1 + m) * n"*)

	       Calc ("op +", eval_binop "#add_"), 
	       Calc ("op *", eval_binop "#mult_"),
	       Calc ("Atools.pow", eval_binop "#power_")
	       ],
      scr = Script ((term_of o the o (parse thy)) scr_expand_binoms)
      }:rls;      


"******* Poly.ML end ******* ...RL";


(**. MG.03: make_polynomial_ ... uses SML-fun for ordering .**)

(*FIXME.0401: make SML-order local to make_polynomial(_) *)
(*FIXME.0401: replace 'make_polynomial'(old) by 'make_polynomial_'(MG) *)
(* Polynom --> List von Monomen *) 
fun poly2list (Const ("op +",_) $ t1 $ t2) = 
    (poly2list t1) @ (poly2list t2)
  | poly2list t = [t];

(* Monom --> Liste von Variablen *)
fun monom2list (Const ("op *",_) $ t1 $ t2) = 
    (monom2list t1) @ (monom2list t2)
  | monom2list t = [t];

(* liefert Variablenname (String) einer Variablen und Basis bei Potenz *)
fun get_basStr (Const ("Atools.pow",_) $ Free (str, _) $ _) = str
  | get_basStr (Free (str, _)) = str
  | get_basStr t = "|||"; (* gross gewichtet; fr Brch ect. *)
(*| get_basStr t = 
    raise error("get_basStr: called with t= "^(term2str t));*)

(* liefert Hochzahl (String) einer Variablen bzw Gewichtstring (zum Sortieren) *)
fun get_potStr (Const ("Atools.pow",_) $ Free _ $ Free (str, _)) = str
  | get_potStr (Const ("Atools.pow",_) $ Free _ $ _ ) = "|||" (* gross gewichtet *)
  | get_potStr (Free (str, _)) = "---" (* keine Hochzahl --> kleinst gewichtet *)
  | get_potStr t = "||||||"; (* gross gewichtet; fr Brch ect. *)
(*| get_potStr t = 
    raise error("get_potStr: called with t= "^(term2str t));*)

(* Umgekehrte string_ord *)
val string_ord_rev =  rev_order o string_ord;
		
 (* Ordnung zum lexikographischen Vergleich zweier Variablen (oder Potenzen) 
    innerhalb eines Monomes:
    - zuerst lexikographisch nach Variablenname 
    - wenn gleich: nach steigender Potenz *)
fun var_ord (a,b: term) = prod_ord string_ord string_ord 
    ((get_basStr a, get_potStr a), (get_basStr b, get_potStr b));

(* Ordnung zum lexikographischen Vergleich zweier Variablen (oder Potenzen); 
   verwendet zum Sortieren von Monomen mittels Gesamtgradordnung:
   - zuerst lexikographisch nach Variablenname 
   - wenn gleich: nach sinkender Potenz*)
fun var_ord_revPow (a,b: term) = prod_ord string_ord string_ord_rev 
    ((get_basStr a, get_potStr a), (get_basStr b, get_potStr b));


(* Ordnet ein Liste von Variablen (und Potenzen) lexikographisch *)
val sort_varList = sort var_ord;

(* Entfernet aeussersten Operator (Wurzel) aus einem Term und schreibt 
   Argumente in eine Liste *)
fun args u : term list =
    let fun stripc (f$t, ts) = stripc (f, t::ts)
	  | stripc (t as Free _, ts) = (t::ts)
	  | stripc (_, ts) = ts
    in stripc (u, []) end;
                                    
(* liefert True, falls der Term (Liste von Termen) nur Zahlen 
   (keine Variablen) enthaelt *)
fun filter_num [] = true
  | filter_num [Free x] = if (is_num (Free x)) then true
				else false
  | filter_num ((Free _)::_) = false
  | filter_num ts =
    (filter_num o (filter_out is_num) o flat o (map args)) ts;

(* liefert True, falls der Term nur Zahlen (keine Variablen) enthaelt 
   dh. er ist ein numerischer Wert und entspricht einem Koeffizienten *)
fun is_nums t = filter_num [t];

(* Berechnet den Gesamtgrad eines Monoms *)
local 
    fun counter (n, []) = n
      | counter (n, x :: xs) = 
	if (is_nums x) then
	    counter (n, xs) 
	else 
	    (case x of 
		 (Const ("Atools.pow", _) $ Free (str_b, _) $ Free (str_h, T)) => 
		     if (is_nums (Free (str_h, T))) then
			 counter (n + (the (int_of_str str_h)), xs)
		     else counter (n + 1000, xs) (*FIXME.MG?!*)
	       | (Const ("Atools.pow", _) $ Free (str_b, _) $ _ ) => 
		     counter (n + 1000, xs) (*FIXME.MG?!*)
	       | (Free (str, _)) => counter (n + 1, xs)
	     (*| _ => raise error("monom_degree: called with factor: "^(term2str x)))*)
	       | _ => counter (n + 10000, xs)) (*FIXME.MG?! ... Brche ect.*)
in  
    fun monom_degree l = counter (0, l) 
end;

(* wie Ordnung dict_ord (lexicographische Ordnung zweier Listen, mit Vergleich 
   der Listen-Elemente mit elem_ord) - Elemente die Bedingung cond erfuellen, 
   werden jedoch dabei ignoriert (uebersprungen)  *)
fun dict_cond_ord _ _ ([], []) = EQUAL
  | dict_cond_ord _ _ ([], _ :: _) = LESS
  | dict_cond_ord _ _ (_ :: _, []) = GREATER
  | dict_cond_ord elem_ord cond (x :: xs, y :: ys) =
    (case (cond x, cond y) of 
	 (false, false) => (case elem_ord (x, y) of 
				EQUAL => dict_cond_ord elem_ord cond (xs, ys) 
			      | ord => ord)
       | (false, true)  => dict_cond_ord elem_ord cond (x :: xs, ys)
       | (true, false)  => dict_cond_ord elem_ord cond (xs, y :: ys)
       | (true, true)  =>  dict_cond_ord elem_ord cond (xs, ys) );

(* Gesamtgradordnung zum Vergleich von Monomen (Liste von Variablen/Potenzen):
   zuerst nach Gesamtgrad, bei gleichem Gesamtgrad lexikographisch ordnen - 
   dabei werden Koeffizienten ignoriert (2*3*a^^^2*4*b gilt wie a^^^2*b) *)
fun degree_ord (xs, ys) =
	    prod_ord int_ord (dict_cond_ord var_ord_revPow is_nums) 
	    ((monom_degree xs, xs), (monom_degree ys, ys));

fun hd_str str = substring (str, 0, 1);
fun tl_str str = substring (str, 1, (size str) - 1);

(* liefert nummerischen Koeffizienten eines Monoms oder None *)
fun get_koeff_of_mon [] =  raise error("get_koeff_of_mon: called with l = []")
  | get_koeff_of_mon (l as x::xs) = if is_nums x then Some x
				    else None;

(* wandelt Koeffizient in (zum sortieren geeigneten) String um *)
fun koeff2ordStr (Some x) = (case x of 
				 (Free (str, T)) => 
				     if (hd_str str) = "-" then (tl_str str)^"0" (* 3 < -3 *)
				     else str
			       | _ => "aaa") (* "num.Ausdruck" --> gross *)
  | koeff2ordStr None = "---"; (* "kein Koeff" --> kleinste *)

(* Order zum Vergleich von Koeffizienten (strings): 
   "kein Koeff" < "0" < "1" < "-1" < "2" < "-2" < ... < "num.Ausdruck" *)
fun compare_koeff_ord (xs, ys) = 
    string_ord ((koeff2ordStr o get_koeff_of_mon) xs,
		(koeff2ordStr o get_koeff_of_mon) ys);

(* Gesamtgradordnung degree_ord + Ordnen nach Koeffizienten falls EQUAL *)
fun koeff_degree_ord (xs, ys) =
	    prod_ord degree_ord compare_koeff_ord ((xs, xs), (ys, ys));

(* Ordnet ein Liste von Monomen (Monom = Liste von Variablen) mittels 
   Gesamtgradordnung *)
val sort_monList = sort koeff_degree_ord;

(* Alternativ zu degree_ord koennte auch die viel einfachere und 
   kuerzere Ordnung simple_ord verwendet werden - ist aber nicht 
   fuer unsere Zwecke geeignet!

fun simple_ord (al,bl: term list) = dict_ord string_ord 
	 (map get_basStr al, map get_basStr bl); 

val sort_monList = sort simple_ord; *)

(* aus 2 Variablen wird eine Summe bzw ein Produkt erzeugt 
   (mit gewuenschtem Typen T) *)
fun plus T = Const ("op +", [T,T] ---> T);
fun mult T = Const ("op *", [T,T] ---> T);
fun binop op_ t1 t2 = op_ $ t1 $ t2;
fun create_prod T (a,b) = binop (mult T) a b;
fun create_sum T (a,b) = binop (plus T) a b;

(* lscht letztes Element einer Liste *)
fun drop_last l = take ((length l)-1,l);

(* Liste von Variablen --> Monom *)
fun create_monom T vl = foldr (create_prod T) (drop_last vl, last_elem vl);
(* Bemerkung: 
   foldr bewirkt rechtslastige Klammerung des Monoms - ist notwendig, damit zwei 
   gleiche Monome zusammengefasst werden knnen (collect_numerals)! 
   zB: 2*(x*(y*z)) + 3*(x*(y*z)) --> (2+3)*(x*(y*z))*)

(* Liste von Monomen --> Polynom *)	
fun create_polynom T ml = foldl (create_sum T) (hd ml, tl ml);
(* Bemerkung: 
   foldl bewirkt linkslastige Klammerung des Polynoms (der Summanten) - 
   bessere Darstellung, da keine Klammern sichtbar! 
   (und discard_parentheses in make_polynomial hat weniger zu tun) *)

(* sorts the variables (faktors) of an expanded polynomial lexicographical *)
fun sort_variables t = 
    let
	val ll =  map monom2list (poly2list t);
	val lls = map sort_varList ll; 
	val T = type_of t;
	val ls = map (create_monom T) lls;
    in create_polynom T ls end;

(* sorts the monoms of an expanded and variable-sorted polynomial 
   by total_degree *)
fun sort_monoms t = 
    let
	val ll =  map monom2list (poly2list t);
	val lls = sort_monList ll;
	val T = type_of t;
	val ls = map (create_monom T) lls;
    in create_polynom T ls end;

(* auch Klammerung muss bereinstimmen; 
   sort_variables klammert Produkte rechtslastig*)
fun is_multUnordered t = ((is_polyexp t) andalso not (t = sort_variables t));

fun eval_is_multUnordered (thmid:string) _ 
		       (t as (Const("Poly.is'_multUnordered", _) $ arg)) thy = 
    if is_multUnordered arg
    then Some (mk_thmid thmid "" 
			((string_of_cterm o cterm_of (sign_of thy)) arg) "", 
	       Trueprop $ (mk_equality (t, HOLogic.true_const)))
    else Some (mk_thmid thmid "" 
			((string_of_cterm o cterm_of (sign_of thy)) arg) "", 
	       Trueprop $ (mk_equality (t, HOLogic.false_const)))
  | eval_is_multUnordered _ _ _ _ = None; 


fun attach_form (_:rule list list) (_:term) (_:term) = (*still missing*)
    []:(rule * (term * term list)) list;
fun init_state (_:term) = e_rrlsstate;
fun locate_rule (_:rule list list) (_:term) (_:rule) =
    ([]:(rule * (term * term list)) list);
fun next_rule (_:rule list list) (_:term) = (None:rule option);
fun normal_form t = Some (sort_variables t,[]:term list);

val order_mult_ =
    Rrls {id = "order_mult_", 
	  prepat = 
	  [([(term_of o the o (parse thy)) "p is_multUnordered"], 
	    (term_of o the o (parse thy)) "?p" )],
	  rew_ord = ("dummy_ord", dummy_ord),
	  erls = append_rls "e_rls-is_multUnordered" e_rls(*MG: poly_erls*)
			    [Calc ("Poly.is'_multUnordered", eval_is_multUnordered "")
			     ],
	  calc = [("plus"    ,("op +"        ,eval_binop "#add_")),
		  ("times"   ,("op *"        ,eval_binop "#mult_")),
		  ("divide_" ,("HOL.divide"  ,eval_cancel "#divide_")),
		  ("power_"  ,("Atools.pow"  ,eval_binop "#power_"))],
	  (*asm_thm=[],*)
	  scr=Rfuns {init_state  = init_state,
		     normal_form = normal_form,
		     locate_rule = locate_rule,
		     next_rule   = next_rule,
		     attach_form = attach_form}};

val order_mult_rls_ = 
  Rls{id = "order_mult_rls_", preconds = [], 
      rew_ord = ("dummy_ord", dummy_ord),
      erls = e_rls,srls = Erls,
      calc = [],
      (*asm_thm = [],*)
      rules = [Rls_ order_mult_
	       ], scr = EmptyScr}:rls;

fun is_addUnordered t = ((is_polyexp t) andalso not (t = sort_monoms t));

(*WN.18.6.03 *)
(*("is_addUnordered", ("Poly.is'_addUnordered", eval_is_addUnordered ""))*)
fun eval_is_addUnordered (thmid:string) _ 
		       (t as (Const("Poly.is'_addUnordered", _) $ arg)) thy = 
    if is_addUnordered arg
    then Some (mk_thmid thmid "" 
			((string_of_cterm o cterm_of (sign_of thy)) arg) "", 
	       Trueprop $ (mk_equality (t, HOLogic.true_const)))
    else Some (mk_thmid thmid "" 
			((string_of_cterm o cterm_of (sign_of thy)) arg) "", 
	       Trueprop $ (mk_equality (t, HOLogic.false_const)))
  | eval_is_addUnordered _ _ _ _ = None; 

fun attach_form (_:rule list list) (_:term) (_:term) = (*still missing*)
    []:(rule * (term * term list)) list;
fun init_state (_:term) = e_rrlsstate;
fun locate_rule (_:rule list list) (_:term) (_:rule) =
    ([]:(rule * (term * term list)) list);
fun next_rule (_:rule list list) (_:term) = (None:rule option);
fun normal_form t = Some (sort_monoms t,[]:term list);

val order_add_ =
    Rrls {id = "order_add_", 
	  prepat = (*WN.18.6.03 Preconditions und Pattern,
		    die beide passen muessen, damit das Rrls angewandt wird*)
	  [([(term_of o the o (parse thy)) "p is_addUnordered"], 
	    (term_of o the o (parse thy)) "?p" 
	    (*WN.18.6.03 also KEIN pattern, dieses erzeugt nur das Environment 
	      fuer die Evaluation der Precondition "p is_addUnordered"*))],
	  rew_ord = ("dummy_ord", dummy_ord),
	  erls = append_rls "e_rls-is_addUnordered" e_rls(*MG: poly_erls*)
			    [Calc ("Poly.is'_addUnordered", eval_is_addUnordered "")
			     (*WN.18.6.03 definiert in Poly.thy,
                               evaluiert prepat*)],
	  calc = [("plus"    ,("op +"        ,eval_binop "#add_")),
		  ("times"   ,("op *"        ,eval_binop "#mult_")),
		  ("divide_" ,("HOL.divide"  ,eval_cancel "#divide_")),
		  ("power_"  ,("Atools.pow"  ,eval_binop "#power_"))],
	  (*asm_thm=[],*)
	  scr=Rfuns {init_state  = init_state,
		     normal_form = normal_form,
		     locate_rule = locate_rule,
		     next_rule   = next_rule,
		     attach_form = attach_form}};

val order_add_rls_ = 
  Rls{id = "order_add_rls_", preconds = [], 
      rew_ord = ("dummy_ord", dummy_ord),
      erls = e_rls,srls = Erls,
      calc = [],
      (*asm_thm = [],*)
      rules = [Rls_ order_add_
	       ], scr = EmptyScr}:rls;

(*. see MG-DA.p.52ff .*)
val make_polynomial(*MG.03, overwrites version from above, 
    previously 'make_polynomial_'*) =
  Seq {id = "make_polynomial", preconds = []:term list, 
      rew_ord = ("dummy_ord", dummy_ord),
      erls = Atools_erls, srls = Erls,calc = [],
      rules = [Rls_ discard_minus_,
	       Rls_ expand_poly_,
	       Calc ("op *", eval_binop "#mult_"),
	       Rls_ order_mult_rls_,
	       Rls_ simplify_power_, 
	       Rls_ calc_add_mult_pow_, 
	       Rls_ reduce_012_mult_,
	       Rls_ order_add_rls_,
	       Rls_ collect_numerals_, 
	       Rls_ reduce_012_,
	       Rls_ discard_parentheses_
	       ],
      scr = EmptyScr
      }:rls;
val norm_Poly(*=make_polynomial*) = 
  Seq {id = "norm_Poly", preconds = []:term list, 
      rew_ord = ("dummy_ord", dummy_ord),
      erls = Atools_erls, srls = Erls, calc = [],
      rules = [Rls_ discard_minus_,
	       Rls_ expand_poly_,
	       Calc ("op *", eval_binop "#mult_"),
	       Rls_ order_mult_rls_,
	       Rls_ simplify_power_, 
	       Rls_ calc_add_mult_pow_, 
	       Rls_ reduce_012_mult_,
	       Rls_ order_add_rls_,
	       Rls_ collect_numerals_, 
	       Rls_ reduce_012_,
	       Rls_ discard_parentheses_
	       ],
      scr = EmptyScr
      }:rls;

(* MG:03 Like make_polynomial_ but without Rls_ discard_parentheses_ 
   and expand_poly_rat_ instead of expand_poly_, see MG-DA.p.56ff*)
(* MG necessary  for termination of norm_Rational(*_mg*) in Rational.ML*)
val make_rat_poly_with_parentheses =
  Seq{id = "make_rat_poly_with_parentheses", preconds = []:term list, 
      rew_ord = ("dummy_ord", dummy_ord),
      erls = Atools_erls, srls = Erls, calc = [],
      rules = [Rls_ discard_minus_,
	       Rls_ expand_poly_rat_,(*ignors rationals*)
	       Calc ("op *", eval_binop "#mult_"),
	       Rls_ order_mult_rls_,
	       Rls_ simplify_power_, 
	       Rls_ calc_add_mult_pow_, 
	       Rls_ reduce_012_mult_,
	       Rls_ order_add_rls_,
	       Rls_ collect_numerals_, 
	       Rls_ reduce_012_
	       (*Rls_ discard_parentheses_ *)
	       ],
      scr = EmptyScr
      }:rls;

(*.a minimal ruleset for reverse rewriting of factions [2];
   compare expand_binoms.*)
val rev_rew_p = 
Seq{id = "reverse_rewriting", preconds = [], rew_ord = ("termlessI",termlessI),
    erls = Atools_erls, srls = Erls,
    calc = [(*("plus"  , ("op +", eval_binop "#add_")), 
	    ("times" , ("op *", eval_binop "#mult_")),
	    ("power_", ("Atools.pow", eval_binop "#power_"))*)
	    ],
    rules = [Thm ("real_plus_binom_times" ,num_str real_plus_binom_times),
	     (*"(a + b)*(a + b) = a ^ 2 + 2 * a * b + b ^ 2*)
	     Thm ("real_plus_binom_times1" ,num_str real_plus_binom_times1),
	     (*"(a +  1*b)*(a + -1*b) = a^^^2 + -1*b^^^2"*)
	     Thm ("real_plus_binom_times2" ,num_str real_plus_binom_times2),
	     (*"(a + -1*b)*(a +  1*b) = a^^^2 + -1*b^^^2"*)

	     Thm ("real_mult_1",num_str real_mult_1),(*"1 * z = z"*)

             Thm ("real_add_mult_distrib" ,num_str real_add_mult_distrib),
	     (*"(z1.0 + z2.0) * w = z1.0 * w + z2.0 * w"*)
	     Thm ("real_add_mult_distrib2",num_str real_add_mult_distrib2),
	     (*"w * (z1.0 + z2.0) = w * z1.0 + w * z2.0"*)
	       
	     Thm ("real_mult_assoc", num_str real_mult_assoc),
	     (*"?z1.1 * ?z2.1 * ?z3. =1 ?z1.1 * (?z2.1 * ?z3.1)"*)
	     Rls_ order_mult_rls_,
	     (*Rls_ order_add_rls_,*)

	     Calc ("op +", eval_binop "#add_"), 
	     Calc ("op *", eval_binop "#mult_"),
	     Calc ("Atools.pow", eval_binop "#power_"),
	     
	     Thm ("sym_realpow_twoI",num_str (realpow_twoI RS sym)),
	     (*"r1 * r1 = r1 ^^^ 2"*)
	     Thm ("sym_real_mult_2",num_str (real_mult_2 RS sym)),
	     (*"z1 + z1 = 2 * z1"*)
	     Thm ("real_mult_2_assoc",num_str real_mult_2_assoc),
	     (*"z1 + (z1 + k) = 2 * z1 + k"*)

	     Thm ("real_num_collect",num_str real_num_collect), 
	     (*"[| l is_const; m is_const |]==>l * n + m * n = (l + m) * n"*)
	     Thm ("real_num_collect_assoc",num_str real_num_collect_assoc),
	     (*"[| l is_const; m is_const |] ==>  
                                     l * n + (m * n + k) =  (l + m) * n + k"*)
	     Thm ("real_one_collect",num_str real_one_collect),
	     (*"m is_const ==> n + m * n = (1 + m) * n"*)
	     Thm ("real_one_collect_assoc",num_str real_one_collect_assoc), 
	     (*"m is_const ==> k + (n + m * n) = k + (1 + m) * n"*)

	     Thm ("realpow_multI", num_str realpow_multI),
	     (*"(r * s) ^^^ n = r ^^^ n * s ^^^ n"*)

	     Calc ("op +", eval_binop "#add_"), 
	     Calc ("op *", eval_binop "#mult_"),
	     Calc ("Atools.pow", eval_binop "#power_"),

	     Thm ("real_mult_1",num_str real_mult_1),(*"1 * z = z"*)
	     Thm ("real_mult_0",num_str real_mult_0),(*"0 * z = 0"*)
	     Thm ("real_add_zero_left",num_str real_add_zero_left)(*0 + z = z*)

	     (*Rls_ order_add_rls_*)
	     ],

    scr = EmptyScr}:rls;      

ruleset' := 
overwritelthy thy (!ruleset',
		   [("norm_Poly", prep_rls norm_Poly),
		    ("Poly_erls",Poly_erls)(*FIXXXME:del with rls.rls'*),
		    ("expand", prep_rls expand),
		    ("expand_poly", prep_rls expand_poly),
		    ("simplify_power", prep_rls simplify_power),
		    ("order_add_mult", prep_rls order_add_mult),
		    ("collect_numerals", prep_rls collect_numerals),
		    ("reduce_012", prep_rls reduce_012),
		    ("discard_parentheses", prep_rls discard_parentheses),
		    ("make_polynomial", prep_rls make_polynomial),
		    ("expand_binoms", prep_rls expand_binoms),
		    ("rev_rew_p", prep_rls rev_rew_p),
		    ("discard_minus_", prep_rls discard_minus_),
		    ("expand_poly_", prep_rls expand_poly_),
		    ("expand_poly_rat_", prep_rls expand_poly_rat_),
		    ("simplify_power_", prep_rls simplify_power_),
		    ("calc_add_mult_pow_", prep_rls calc_add_mult_pow_),
		    ("reduce_012_mult_", prep_rls reduce_012_mult_),
		    ("collect_numerals_", prep_rls collect_numerals_),
		    ("reduce_012_", prep_rls reduce_012_),
		    ("discard_parentheses_",prep_rls discard_parentheses_),
		    ("order_mult_rls_", prep_rls order_mult_rls_),
		    ("order_add_rls_", prep_rls order_add_rls_),
		    ("make_rat_poly_with_parentheses", 
		     prep_rls make_rat_poly_with_parentheses)
		    (*("", prep_rls ),
		     ("", prep_rls ),
		     ("", prep_rls )
		     *)
		    ]);

calclist':= overwritel (!calclist', 
   [("is_polyrat_in", ("Poly.is'_polyrat'_in", 
		       eval_is_polyrat_in "#eval_is_polyrat_in")),
    ("is_expanded_in", ("Poly.is'_expanded'_in", eval_is_expanded_in "")),
    ("is_poly_in", ("Poly.is'_poly'_in", eval_is_poly_in "")),
    ("has_degree_in", ("Poly.has'_degree'_in", eval_has_degree_in "")),
    ("is_polyexp", ("Poly.is'_polyexp", eval_is_polyexp "")),
    ("is_multUnordered", ("Poly.is'_multUnordered", eval_is_multUnordered"")),
    ("is_addUnordered", ("Poly.is'_addUnordered", eval_is_addUnordered ""))
    ]);

(** problems **)

store_pbt
 (prep_pbt Poly.thy "pbl_simp_poly" [] e_pblID
 (["polynomial","simplification"],
  [("#Given" ,["term t_"]),
   ("#Where" ,["t_ is_polyexp"]),
   ("#Find"  ,["normalform n_"])
  ],
  append_rls "e_rls" e_rls [(*for preds in where_*)
			    Calc ("Poly.is'_polyexp", eval_is_polyexp "")], 
  Some "Simplify t_", 
  [["simplification","for_polynomials"]]));

(** methods **)

store_met
    (prep_met Poly.thy "met_simp_poly" [] e_metID
	      (["simplification","for_polynomials"],
	       [("#Given" ,["term t_"]),
		("#Where" ,["t_ is_polyexp"]),
		("#Find"  ,["normalform n_"])
		],
	       {rew_ord'="tless_true",
		rls' = e_rls,
		calc = [], 
		srls = e_rls, 
		prls = append_rls "simplification_for_polynomials_prls" e_rls 
				  [(*for preds in where_*)
				   Calc ("Poly.is'_polyexp",eval_is_polyexp"")],
		crls = e_rls, nrls = e_rls},
	       "Script SimplifyScript (t_::real) =                \
	       \  ((Rewrite_Set norm_Poly False) t_)"
	       ));
